#include "CTextureLoader_GAF.h"
#include "CFileManager.h"
//#include "CConstantManager.h"
#include "ITA_Headers.h"
#include <fstream>
#include <algorithm>
//#include "CDialogBox.h"
#include "CTextParser.h"
#include "CTextureFactory.h"
#include "CPaletteLoader_PAL.h"
#include <functional>

int CHANNEL = 3;

std::vector<CTextureLoader_GAF::Texture*> CTextureLoader_GAF::m_TextureArray;
std::string CTextureLoader_GAF::m_CurrentFile;
std::string CTextureLoader_GAF::m_SaveDir;
std::string CTextureLoader_GAF::m_SaveFile;

struct isEqualGAF : public std::unary_function<CTextureLoader_GAF::Texture*, bool>
{
	std::string m_Name;

	isEqualGAF(std::string Name)
	{
		m_Name = Name;
	};
	bool operator()(CTextureLoader_GAF::Texture* x) 
	{
		if (m_Name == x->m_Name)
			return true;

		return false;
	}
};


struct isLessGAF : public std::binary_function<CTextureLoader_GAF::Texture*, CTextureLoader_GAF::Texture*, bool>
{
	isLessGAF()
	{};
	bool operator()(CTextureLoader_GAF::Texture *a, CTextureLoader_GAF::Texture *b) 
	{
		if (a->m_sizeY < b->m_sizeY)
			return true;

		return false;
	}
};

bool CTextureLoader_GAF::CreateTextureList(std::vector<std::string> TextureList, std::string DirName, std::string SaveDir, std::string UnitName, bool isTextured)
{
	if (TextureList.size() <=0 || DirName == "" || isTextured == false)
		return false;

	m_SaveDir = SaveDir;
	m_SaveFile = UnitName;

	DirName = CTextParser::TextReplace( DirName, "\\", "/");

	if (DirName[DirName.length() - 1 ] != '/')
		DirName += "/";

	std::vector<std::string> FileList = CFileManager::FindAllFiles(DirName);

	//cPlugIn::Msgbox();

	for (int i = 0 ; i < FileList.size() ; i++)
	{
		std::transform(FileList[i].begin(), FileList[i].end(), FileList[i].begin(), toupper);

		if (FileList[i].substr( FileList[i].length() - 4, FileList[i].length() ) == ".GAF" )
		{

		std::ifstream inputFile( (DirName + FileList[i]).c_str(), std::ios::in | std::ios::binary);
		if ( inputFile.fail())
		{
			return false;	// "Couldn't open the model file."
		}

		m_CurrentFile = FileList[i];

		inputFile.seekg( 0, std::ios::end );
		long fileSize = inputFile.tellg() ;
		inputFile.seekg( 0, std::ios::beg );

		unsigned char *pBuffer = new unsigned char[fileSize];
		inputFile.read( (char*)pBuffer, fileSize );
		inputFile.close();

		const unsigned char *pPtr = pBuffer;
		GAFHEADER *pHeader = ( GAFHEADER* )pPtr;
		pPtr += sizeof( GAFHEADER );


		
		if (pHeader->IDVersion == 0x00010100)
		{
			long *pList = ( long* )pPtr;
			for (int j = 0 ; j < pHeader->Entries ; j++)
			{
				GAFENTRY *pEntry = ( GAFENTRY* )(pBuffer + pList[j]);

				std::vector<std::string>::iterator result = std::find(TextureList.begin(), TextureList.end(), pEntry->Name);

				static int pos = 0;

				//if (strcmp("Frond01",pEntry->Name) == 0)
				if ( result != TextureList.end() )
				{
					if(pEntry->Frames > 0)
					{
						m_TextureArray.push_back( GetTexture( (pBuffer + pList[j] + sizeof(GAFENTRY)), pBuffer) );
						m_TextureArray[m_TextureArray.size() - 1]->m_Name = *result;
						//m_TextureArray[m_TextureArray.size() - 1].m_Name = *result + ".png";
						//CTextureFactory::SaveTextureUnit(TextureArray[pos].m_Name, TextureArray[pos].m_data, TextureArray[pos].m_sizeX, TextureArray[pos].m_sizeY);
						//pos++;
					}
					//else
					//	CDialogBox::WARNING("The file " + m_CurrentFile + " contains an illegal texture entry that the model is trying to use. The texture name is " + *result, "CTextureLoader_GAF : ERROR");
				}

			}
		}

		delete [] pBuffer;
		delete pHeader;
		}
	}

	if (m_TextureArray.size() <= 0)
		return false;

	std::sort( m_TextureArray.begin(), m_TextureArray.end(), isLessGAF() );

	GetConsolidatedFrameData(m_TextureArray);

	for(std::vector<Texture*>::iterator itor = m_TextureArray.begin() ; itor != m_TextureArray.end() ; itor++)
	{
		if ((*itor) == NULL)
		{
			delete (*itor)->m_data;
			delete (*itor);
		}
	}

//	CPaletteLoader_PAL::shutdownPalette();

	return true;
}

CTextureLoader_GAF::Texture *CTextureLoader_GAF::GetTexture(unsigned char *EntryBase, unsigned char *Base)
{
	GAFFRAMEENTRY *pFrameEntry = ( GAFFRAMEENTRY* )EntryBase;
	GAFFRAMEDATA *pFrameData = ( GAFFRAMEDATA* )(Base + pFrameEntry->PtrFrameTable);

	Texture *Output;

	if(pFrameData->FramePointers == 0)
	{
		if (pFrameData->Compressed == true)
		{
			Output = GetCompressedFrameData(Base + pFrameData->PtrFrameData, pFrameData->Width, pFrameData->Height);
		}
		else
		{
			Output = GetUnompressedFrameData(Base + pFrameData->PtrFrameData, pFrameData->Width, pFrameData->Height);
		}
	}
	else
	{

//		CDialogBox::WARNING("The file " + m_CurrentFile + " contains an illegal texture entry that the model is trying to use. The texture name is ", "CTextureLoader_GAF : ERROR");
	}

	ConvertTo24Bit(Output);

//	CTextureFactory::SaveTextureUnit("Hi.tga", Output.m_data, Output.m_sizeX, Output.m_sizeY);
	return Output;
}


CTextureLoader_GAF::Texture *CTextureLoader_GAF::GetCompressedFrameData(unsigned char *FramePtr, int sizeX, int sizeY)
{
	Texture *Output = new Texture;
	return Output;
}
CTextureLoader_GAF::Texture *CTextureLoader_GAF::GetUnompressedFrameData(unsigned char *FramePtr, int sizeX, int sizeY)
{
	unsigned char *OriginalData =  ( unsigned char* )FramePtr;
	Texture *Output = new Texture;

	Output->m_sizeX = sizeX;
	Output->m_sizeY = sizeY;
	Output->m_data = new unsigned char[sizeX * sizeY];
	memcpy(Output->m_data, OriginalData, sizeof(unsigned char) * sizeX * sizeY);

	return Output;
}

void CTextureLoader_GAF::ConvertTo24Bit(Texture *&Original)
{
	unsigned char *NewData = new unsigned char[Original->m_sizeX * Original->m_sizeY * CHANNEL];
	
	for(int i = 0 ; i < (Original->m_sizeX * Original->m_sizeY) ; i++)
	{
		memcpy(&NewData[i * CHANNEL], CPaletteLoader_PAL::getIndexRgbaValue(Original->m_data[i]), sizeof(unsigned char) * CHANNEL);
	}	

	delete [] Original->m_data;
	Original->m_data = NewData;
}

void CTextureLoader_GAF::GetConsolidatedFrameData(std::vector<Texture*> &TextureArray)
{
	int sizeX = 16,
		sizeY = 16,
		currentSizeX = 0,
		currentSizeY = 0,
		spacingY = 0;

	bool isTooSmall = true;

	while(isTooSmall)
	{
		isTooSmall = false;
		spacingY = TextureArray[TextureArray.size() - 1]->m_sizeY;
		for (int i = TextureArray.size() - 1 ; i >= 0 ; i--)
		{
			currentSizeX += TextureArray[i]->m_sizeX;

			
			if (i - 1 >= 0)
			{
				if (currentSizeX + TextureArray[i - 1]->m_sizeX >= sizeX)
				{

					//i--;
					currentSizeX = 0;
					currentSizeY += spacingY;
					if (currentSizeY + TextureArray[i - 1]->m_sizeY >= sizeY)
					{
						sizeX = sizeX * 2;
						sizeY = sizeY * 2;
						currentSizeX = 0;
						currentSizeY = 0;
						spacingY = 0;
						isTooSmall = true;

						i = -1;
					}

					if(i + 1 < TextureArray.size())
					{
						spacingY = TextureArray[i+1]->m_sizeY;
					}

				}
			}
		}
	}
	
	unsigned char *ConsolidatedFrame = new unsigned char[sizeX * sizeY * CHANNEL];
	for (int i = 0 ; i < sizeX * sizeY * CHANNEL ; i++)
		ConsolidatedFrame[i] = 1;

	currentSizeX = 0;
	currentSizeY = 0;
	spacingY = 0;

	spacingY = TextureArray[TextureArray.size() - 1]->m_sizeY;
	for (i = TextureArray.size() - 1 ; i >= 0 ; i--)
	{
		for(int j = 0 ; j < TextureArray[i]->m_sizeY ; j++)
		{
			memcpy(&ConsolidatedFrame[( (currentSizeY + j) * sizeX + currentSizeX ) * CHANNEL], &TextureArray[i]->m_data[j * TextureArray[i]->m_sizeX * CHANNEL], sizeof(char) * TextureArray[i]->m_sizeX * CHANNEL);
		}

		TextureArray[i]->m_TexCoord[0][0] = (float)currentSizeX / sizeX;
		TextureArray[i]->m_TexCoord[0][1] = 1 - ((float)currentSizeY / sizeY);

		TextureArray[i]->m_TexCoord[3][0] = (float)currentSizeX / sizeX;;
		TextureArray[i]->m_TexCoord[3][1] = 1 - ((float)(currentSizeY + TextureArray[i]->m_sizeY)/ sizeY);

		TextureArray[i]->m_TexCoord[2][0] = ((float)(currentSizeX + TextureArray[i]->m_sizeX) / sizeX);
		TextureArray[i]->m_TexCoord[2][1] = 1 - ((float)(currentSizeY + TextureArray[i]->m_sizeY) / sizeY);

		TextureArray[i]->m_TexCoord[1][0] = ((float)(currentSizeX + TextureArray[i]->m_sizeX) / sizeX);
		TextureArray[i]->m_TexCoord[1][1] = 1 - ((float)currentSizeY / sizeY);

		currentSizeX += TextureArray[i]->m_sizeX;

		if (i - 1 >= 0)
		{
			if (currentSizeX + TextureArray[i - 1]->m_sizeX >= sizeX)
			{
				currentSizeX = 0;
				currentSizeY += spacingY;

				if(i + 1 < TextureArray.size())
				{
					spacingY = TextureArray[i+1]->m_sizeY;
				}
			}
		}
	}


	//cPlugIn::Msgbox(m_SaveDir + "Main.png");
	//CTextureFactory::SaveTextureUnit("data/Texture/Main.png", ConsolidatedFrame, sizeX, sizeY);


	//cPlugIn::Msgbox("HI");

	CTextureFactory::SaveTextureUnit(m_SaveDir + m_SaveFile, ConsolidatedFrame, sizeX, sizeY);
	delete [] ConsolidatedFrame;
}

void CTextureLoader_GAF::getTextureCoordinates(std::string textureName, int index, float &u, float &v)
{
	std::vector<Texture*>::iterator result = std::find_if(m_TextureArray.begin(), 
														 m_TextureArray.end(), 
													     isEqualGAF(textureName));

	if (result != m_TextureArray.end() )
	{
		u = (*result)->m_TexCoord[index][0];
		v = (*result)->m_TexCoord[index][1];
	}
	else
	{
		u = 0;
		v = 0;
	}
}

bool CTextureLoader_GAF::Texture::operator < ( const CTextureLoader_GAF::Texture& in )
{
	if (m_sizeY  < in.m_sizeY)
		return true;

	return false;
}

bool CTextureLoader_GAF::Texture::operator == ( const Texture& in )
{
	if (m_Name == in.m_Name)
		return true;

	return false;
}

CTextureLoader_GAF::Texture::~Texture()
{
	delete [] m_data;
}
